# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
import tensorflow as tf
import tensorflow_probability as tfp
from scipy import linalg
from scipy.special import gegenbauer as ScipyGegenbauer

from spherical_harmonics.fundamental_set import (
    FundamentalSystemCache,
    calculate_decrement_in_determinant,
    grad_calculate_decrement_in_determinant,
)
from spherical_harmonics.gegenbauer_polynomial import Gegenbauer


def det(Z, X_system, C_old, gegenbauer: Gegenbauer):
    """
    Objective to

    Z: [d]
    X_system: [N, d]
    C_old: [N, N]
    """
    norm = tf.reduce_sum(Z**2) ** 0.5  # scalar
    X = Z / norm  # [d]
    XXd = tf.einsum("nd,d->n", X_system, X)  # [N]
    M_cross = gegenbauer(XXd)  # [N]

    # Determinant is given by the compuations below though we only compute the
    # bits that depend on Z and make sure the objective can be minimised in
    # order to maximise the overall determinant:
    # C_1 = gegenbauer.value_at_1
    # det_C_old = tf.linalg.det(C_old)
    # return det_C_old * (C_1 - tf.reduce_sum(res ** 2))

    C_old_chol = tf.linalg.cholesky(C_old)
    res = tf.linalg.triangular_solve(C_old_chol, M_cross[:, None], lower=True)
    return tf.reduce_sum(res**2)


@pytest.mark.parametrize("dimension", [3, 5, 6, 9])
@pytest.mark.parametrize("degree", [5, 4, 5])
def test_objective(dimension, degree):
    alpha = (dimension - 2) / 2
    gegenbauer = ScipyGegenbauer(degree, alpha)
    system = FundamentalSystemCache(dimension=dimension)
    X_system = system.load(degree)
    C_new = gegenbauer(X_system @ X_system.T)
    Z = X_system[-1]
    X_system = X_system[:-1]
    C_old = gegenbauer(X_system @ X_system.T)
    det1 = linalg.det(C_new)

    C_1 = gegenbauer(1.0)
    det_C_old = tf.linalg.det(C_old)
    v = calculate_decrement_in_determinant(
        Z, X_system, linalg.cholesky(C_old, lower=True), gegenbauer
    )
    det2 = det_C_old * (C_1 - v)
    np.testing.assert_allclose(det1, det2)


@pytest.mark.parametrize("dimension", [3, 5, 6, 9])
@pytest.mark.parametrize("degree", [3, 4, 5])
def test_grad_objective(dimension, degree):
    alpha = (dimension - 2) / 2
    gegenbauer = Gegenbauer(degree, alpha)
    system = FundamentalSystemCache(dimension=dimension)
    X_system = system.load(degree)
    X_system = tf.convert_to_tensor(X_system, dtype=tf.float64)
    Z = tf.random.normal((dimension,), dtype=tf.float64)
    X_system = X_system[:-1]
    C_old = gegenbauer(tf.matmul(X_system, X_system, transpose_b=True))

    _, dv1 = tfp.math.value_and_gradient(
        lambda Z: det(Z, X_system, C_old, gegenbauer), Z
    )
    dv1 = dv1.numpy()
    dv2 = grad_calculate_decrement_in_determinant(
        Z.numpy(),
        X_system.numpy(),
        linalg.cholesky(C_old.numpy(), lower=True),
        ScipyGegenbauer(degree, alpha),
    )
    np.testing.assert_array_almost_equal(dv1, dv2)



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import importlib.resources
import warnings
from typing import Optional

import numpy as np
from scipy import linalg, optimize
from scipy.special import comb as combinations
from scipy.special import gegenbauer as ScipyGegenbauer


def num_harmonics(dimension: int, degree: int) -> int:
    r"""
    Number of spherical harmonics of a particular degree n in
    d dimensions. Referred to as N(d, n).

    param dimension:
        S^{d-1} = { x ∈ R^d and ||x||_2 = 1 }
        For a circle d=2, for a ball d=3
    param degree: degree of the harmonic
    """
    if degree == 0:
        return 1
    elif dimension == 3:
        return int(2 * degree + 1)
    else:
        comb = combinations(degree + dimension - 3, degree - 1)
        return int(np.round((2 * degree + dimension - 2) * comb / degree))


class FundamentalSystemCache:
    """A simple cache object to access precomputed fundamental system.

    Fundamental system are sets of points that allow the user to evaluate the spherical
    harmonics in an arbitrary dimension"""

    def __init__(
        self,
        dimension: int,
        load_dir="fundamental_system",
        only_use_cache: bool = True,
        strict_loading: bool = True,
    ):
        self.file_name = f"{load_dir}/fs_{dimension}D.npz"
        self.resource_file = (
            importlib.resources.files("spherical_harmonics") / self.file_name
        )
        self.dimension = dimension
        self.only_use_cache = only_use_cache
        self.strict_loading = strict_loading

        if self.resource_file.is_file():
            with self.resource_file.open("rb") as f:
                with np.load(f) as data:
                    self.cache = {k: v for (k, v) in data.items()}
        elif only_use_cache and self.strict_loading:
            raise ValueError(
                f"Fundamental system for dimension {dimension} has not been precomputed."
                "Terminating computations. Precompute set by running `fundamental_set.py`"
            )
        else:
            self.cache = {}

    @property
    def max_computed_degree(self) -> int:
        max_degree = -1
        while True:
            max_degree = max_degree + 1
            key = self.cache_key(max_degree)
            if key not in self.cache:
                break
        return max_degree

    def cache_key(self, degree: int) -> str:
        """Return the key used in the cache"""
        return f"degree_{degree}"

    def load(self, degree: int) -> Optional[np.ndarray]:
        """Load or calculate the set for given degree"""
        key = self.cache_key(degree)
        if key not in self.cache:
            if self.only_use_cache and self.strict_loading:
                raise ValueError(
                    f"Fundamental system for dimension {self.dimension} and degree "
                    f"{degree} has not been precomputed. Terminating "
                    "computations. Precompute set by running `fundamental_set.py`"
                )
            elif not self.only_use_cache:
                print("WARNING: Cache miss - calculating  system")
                self.cache[key] = self.calculate(self.dimension, degree)

        if key not in self.cache and not self.strict_loading:
            warnings.warn(
                f"Fundamental system for dimension {self.dimension} has been "
                f"computed up to degree {self.max_computed_degree}. This means you will not "
                "be able to evaluate individual spherical harmonics for larger degrees.",
                RuntimeWarning,
            )
            return None

        return self.cache[key]

    def regenerate_and_save_cache(self, max_degrees: int) -> None:
        """Regenerate and overwrite saved cache to disk"""
        system = {}
        for degree in range(max_degrees):
            print(f"finding level {degree}/{max_degrees} in {self.dimension}D")
            d_system = self.calculate(
                self.dimension, degree, gtol=1e-8, num_restarts=10
            )
            system[f"degree_{degree}"] = d_system
        with open(self.file_name, "wb+") as f:
            np.savez(f, allow_pickle=True, **system)

    @staticmethod
    def calculate(
        dimension: int, degree: int, *, gtol: float = 1e-5, num_restarts: int = 1
    ) -> np.ndarray:
        return build_fundamental_system(
            dimension, degree, gtol=gtol, num_restarts=num_restarts
        )


def build_fundamental_system(
    dimension, degree, *, gtol=1e-5, num_restarts=1, debug=False
):
    """
    We build a fundamental system incrementally, by adding a new candidate vector each
    time and maximising the span of the space generated by these spherical harmonics.

    This can be done by greedily optimising the determinant of the gegenbauered Gram matrix.

    Based on [1, Defintion 3.1]

    [1] Approximation Theory and Harmonic Analysis on Spheres and Balls,
        Feng Dai and Yuan Xu, Chapter 1. Spherical Harmonics,
        https://arxiv.org/pdf/1304.2585.pdf
    """
    alpha = (dimension - 2) / 2.0
    gegenbauer = ScipyGegenbauer(degree, alpha)
    system_size = num_harmonics(dimension, degree)

    # 1. Choose first direction in system to be north pole
    Z0 = np.eye(dimension)[-1]
    X_system = normalize(Z0).reshape(1, dimension)
    M_system_chol = cholesky_of_gegenbauered_gram(gegenbauer, X_system)

    # 2. Find a new vector incrementally by max'ing the determinant of the gegenbauered Gram
    for i in range(1, system_size):

        Z_next, ndet, restarts = None, np.inf, 0
        while restarts <= num_restarts:
            x_init = np.random.randn(dimension)
            result = optimize.fmin_bfgs(
                f=calculate_decrement_in_determinant,
                fprime=grad_calculate_decrement_in_determinant,
                x0=x_init,
                args=(X_system, M_system_chol, gegenbauer),
                full_output=True,
                gtol=gtol,
                disp=debug,
            )

            if result[1] <= ndet:
                Z_next, ndet, *_ = result
                #  TODO: we should we break when we find the best vector.
                #  Unclear how to do this at this point.
            # Try again with new x_init
            restarts += 1
        print(
            f"det: {-ndet:11.4f}, ({i + 1:5d} of {system_size:5d}, "
            f"degree {degree}: {dimension}D)"
        )
        X_next = normalize(Z_next).reshape(1, dimension)
        X_system = np.vstack([X_system, X_next])
        M_system_chol = cholesky_of_gegenbauered_gram(gegenbauer, X_system)

    return X_system


def calculate_decrement_in_determinant(Z, X_system, M_system_chol, gegenbauer):
    r"""Calculate the negative determinant.

    :param Z: is a potential vector for the next fundamental point (it will get normalized)
    :param X_system: is a matrix of existing fundamental points [num_done, D]
    :param M_system_chol: is the cholesky of the matrix M of the done points [num_done, num_done]

    :return: the negative-increment of the determinant of the matrix with Z (normalized)
     added to the done points
    """
    X = normalize(Z)
    XXd = np.dot(X_system, X)  # [num_done,]

    # M_new = gegenbauer(1.0)  # X normalized so X @ X^T = 1
    M_cross = gegenbauer(XXd)

    # Determinant of M is computed efficiently making use of the Schur complement
    # M = [[ M_system_chol, M_cross], [ M_cross^T, M_new]]
    # det(M) = det(M_system_chol) * det(M_new - M_cross^T M_system_chol^{-1} M_cross )
    res = linalg.solve_triangular(M_system_chol, M_cross, trans=0, lower=True)
    return np.sum(np.square(res))


def grad_calculate_decrement_in_determinant(Z, X_system, M_system_chol, gegenbauer):
    r"""Calculate the negative determinant.

    :param Z: is a potential vector for the next fundamental point (it will get normalized)
    :param X_system: is a matrix of existing fundamental points [num_done, D]
    :param M_system_chol: is the cholesky of the matrix M of the done points [num_done, num_done]

    """
    X = normalize(Z)
    XXd = np.dot(X_system, X)  # [num_done,]

    M_cross = gegenbauer(XXd)

    res = linalg.solve_triangular(M_system_chol, M_cross, trans=0, lower=True)
    dM_cross = 2.0 * linalg.solve_triangular(
        M_system_chol,
        res,
        trans=1,
        lower=True,
    )
    dXXd = gegenbauer.deriv()(XXd) * dM_cross
    dX = np.dot(X_system.T, dXXd)
    dZ = (dX - X * np.dot(X, dX)) / norm(Z)
    return dZ


def cholesky_of_gegenbauered_gram(gegenbauer_polynomial, x_matrix):
    XtX = x_matrix @ x_matrix.T
    return np.linalg.cholesky(gegenbauer_polynomial(XtX))


def normalize(vec: np.ndarray):
    assert len(vec.shape) == 1
    return vec / norm(vec)


def norm(vec: np.ndarray):
    assert len(vec.shape) == 1
    return np.sqrt(np.sum(np.square(vec)))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pre-calculate fundamental system")
    parser.add_argument("-d", "--dim", default=3, type=int, help="Dimension")
    degrees_levels_group = parser.add_mutually_exclusive_group()
    degrees_levels_group.add_argument("-m", "--max-harmonics", default=1000, type=int)
    degrees_levels_group.add_argument("-l", "--max-degrees", type=int)

    args = parser.parse_args()

    def calc_degrees(dimension: int, max_harmonics: int):
        harmonics = 0
        degree = 1
        while harmonics < max_harmonics:
            harmonics += num_harmonics(dimension, degree)
            degree += 1
        degree -= 1
        return degree

    def regenerate_cache(dimension: int, max_degrees: int):
        FundamentalSystemCache(
            dimension, only_use_cache=False
        ).regenerate_and_save_cache(max_degrees)

    max_degrees = args.max_degrees or calc_degrees(args.dim, args.max_harmonics)

    regenerate_cache(args.dim, max_degrees)



import lab as B
from beartype.typing import List
from lab import dispatch
from lab.util import abstract
from plum import Union


@dispatch
@abstract()
def polyval(coeffs: list, x: B.Numeric):
    """
    Computes the elementwise value of a polynomial.

    If `x` is a tensor and `coeffs` is a list if size n + 1, this function returns
    the value of the n-th order polynomial

    ..math:
        p(x) = coeffs[n-1] + coeffs[n-2] * x + ... + coeffs[0] * x**(n-1)
    """


@dispatch
@abstract()
def from_numpy(_: B.Numeric, b: Union[list, List, B.Numeric, B.NPNumeric]):
    """
    Converts the array `b` to a tensor of the same backend as `a`
    """



import lab as B
import tensorflow as tf
from beartype.typing import List
from lab import dispatch
from plum import Union

_Numeric = Union[B.Number, B.TFNumeric, B.NPNumeric]


@dispatch
def polyval(coeffs: list, x: _Numeric) -> _Numeric:  # type: ignore
    """
    Computes the elementwise value of a polynomial.

    If `x` is a tensor and `coeffs` is a list if size n + 1, this function returns
    the value of the n-th order polynomial

    ..math:
        p(x) = coeffs[n-1] + coeffs[n-2] * x + ... + coeffs[0] * x**(n-1)
    """
    coeff_list = [B.cast(B.dtype(x), coeff) for coeff in coeffs]
    return tf.math.polyval(coeff_list, x)


@dispatch
def from_numpy(a: B.TFNumeric, b: Union[list, List, B.Numeric, B.NPNumeric, B.TFNumeric]):  # type: ignore
    """
    Converts the array `b` to a tensor of the same backend as `a`
    """
    return tf.convert_to_tensor(b)



from typing import List, Union

import lab as B
import numpy as np
import pytest
import tensorflow as tf

import spherical_harmonics.tensorflow  # noqa
from spherical_harmonics.fundamental_set import (
    FundamentalSystemCache,
    build_fundamental_system,
)
from spherical_harmonics.gegenbauer_polynomial import GegenbauerManualCoefficients
from spherical_harmonics.spherical_harmonics import (
    SphericalHarmonics,
    SphericalHarmonicsLevel,
)
from spherical_harmonics.utils import (
    spherical_to_cartesian,
    spherical_to_cartesian_4d,
    surface_area_sphere,
)


@pytest.mark.parametrize("max_degree", list(range(1, 10, 3)) + [34])
def test_orthonormal_basis_3d(max_degree):
    """Numerical check that int_{S^2} Y_i(x) Y_j(x) dx = dirac(i,j)"""

    num_grid = 300
    dimension = 3

    theta = np.linspace(0, 2 * np.pi, num_grid)  # [N]
    phi = np.linspace(0, np.pi, num_grid)  # [N]
    theta, phi = np.meshgrid(theta, phi)  # [N, N], [N, N]
    x_spherical = np.c_[theta.reshape(-1, 1), phi.reshape(-1, 1)]  # [N^2, 2]
    x_cart = spherical_to_cartesian(x_spherical)

    harmonics = SphericalHarmonics(dimension, max_degree)
    harmonics_at_x = tf.transpose(harmonics(x_cart)).numpy()  # [M, N^2]

    d_x_spherical = 2 * np.pi**2 / num_grid**2
    inner_products = (
        harmonics_at_x
        # sin(phi) to account for the surface area element of S^2
        @ (harmonics_at_x.T * np.sin(x_spherical[:, [1]]))
        * d_x_spherical
    )

    inner_products = inner_products / surface_area_sphere(dimension)

    np.testing.assert_array_almost_equal(
        inner_products, np.eye(len(harmonics_at_x)), decimal=1
    )


@pytest.mark.parametrize("max_degree", range(1, 8, 3))
def test_orthonormal_basis_4d(max_degree):
    """Numerical check that int_{S^3} Y_i(x) Y_j(x) dx = dirac(i,j)"""
    num_grid = 25
    dimension = 4

    theta1 = np.linspace(0, 2 * np.pi, num_grid)
    theta2 = np.linspace(0, np.pi, num_grid)
    theta3 = np.linspace(0, np.pi, num_grid)
    theta1, theta2, theta3 = np.meshgrid(theta1, theta2, theta3)
    x_spherical = np.c_[
        theta1.reshape((-1, 1)),
        theta2.reshape((-1, 1)),
        theta3.reshape((-1, 1)),
    ]  # [N^3, 3]
    x_cart = spherical_to_cartesian_4d(x_spherical)

    harmonics = SphericalHarmonics(dimension, max_degree)
    harmonics_at_x = tf.transpose(harmonics(x_cart))  # [M, N^3]

    d_x_spherical = 2 * np.pi**3 / num_grid**3

    inner_products = np.ones((len(harmonics_at_x), len(harmonics_at_x))) * np.nan

    for i, Y1 in enumerate(harmonics_at_x):
        for j, Y2 in enumerate(harmonics_at_x):
            v = np.sum(
                Y1
                * Y2
                # account for surface area element of S^3 sphere
                * np.sin(x_spherical[:, -1]) ** 2
                * np.sin(x_spherical[:, -2])
                * d_x_spherical
            )
            inner_products[i, j] = v

    inner_products = inner_products / surface_area_sphere(dimension)

    np.testing.assert_array_almost_equal(
        inner_products, np.eye(len(harmonics)), decimal=1
    )


@pytest.mark.parametrize("dimension", range(3, 11, 3))
@pytest.mark.parametrize("max_degree", range(2, 7, 3))
def test_equality_spherical_harmonics_collections(dimension, max_degree):

    fast_harmonics = SphericalHarmonics2(dimension, max_degree)
    harmonics = SphericalHarmonics(dimension, max_degree)

    num_points = 100
    X = np.random.randn(num_points, dimension)
    # make unit vectors
    X /= np.sum(X**2, axis=-1, keepdims=True) ** 0.5

    np.testing.assert_array_almost_equal(
        fast_harmonics(X),
        harmonics(X),
    )


@pytest.mark.parametrize("dimension", range(3, 7, 1))
@pytest.mark.parametrize("degree", range(1, 7, 3))
def test_addition_theorem(dimension, degree):
    fundamental_system = FundamentalSystemCache(dimension)
    harmonics = SphericalHarmonicsLevel(dimension, degree, fundamental_system)
    X = np.random.randn(100, dimension)
    X = X / (np.sum(X**2, keepdims=True, axis=1) ** 0.5)
    harmonics_at_X = harmonics(X)[..., None]  # [M:=N(dimension, degree), N, 1]
    harmonics_xxT = tf.matmul(
        harmonics_at_X, harmonics_at_X, transpose_b=True
    )  # [M, N, N]

    # sum over all harmonics in the level
    # addition_manual = harmonics_at_X.T @ harmonics_at_X  # [N, N]
    addition_manual = tf.reduce_sum(harmonics_xxT, axis=0)  # [N, N]
    addition_theorem = harmonics.addition(X)

    np.testing.assert_array_almost_equal(addition_manual, addition_theorem)

    np.testing.assert_array_almost_equal(
        np.diag(addition_manual)[..., None], harmonics.addition_at_1(X)
    )


@pytest.mark.parametrize("dimension", range(3, 21))
def test_init_spherical_harmonics(dimension):
    max_degree = 2
    _ = SphericalHarmonics(dimension, max_degree)


@pytest.mark.parametrize("dimension", range(21, 25))
def test_init_spherical_harmonics_not_cached(dimension):
    max_degree = 2
    with pytest.raises(ValueError):
        _ = SphericalHarmonics(dimension, max_degree)


def test_building_fundamental_set_shapes():
    dimension = 5
    x_system1 = build_fundamental_system(dimension, degree=1)
    assert x_system1.shape == (5, 5)
    x_system2 = build_fundamental_system(dimension, degree=0)
    assert x_system2.shape == (1, 5)


class SphericalHarmonics2(SphericalHarmonics):
    """
    Slightly faster implementation (approx 10%) of the `__call__` method than
    the one in `SphericalHarmonicsCollection` as we don't make use of a `map`.
    """

    def __init__(
        self, dimension: int, degrees: Union[int, List[int]], debug: bool = True
    ):
        """
        :param dimension: if d = dimension, then
            S^{d-1} = { x ∈ R^d and ||x||_2 = 1 }
            For a circle d=2, for a ball d=3
        :param degrees: list of degrees of spherical harmonic levels,
            if integer all levels (or degrees) up to `degrees` are used.
        :param debug: print debug messages.
        """
        super().__init__(dimension, degrees, debug)

        # Hack: overwrite levels to use `GegenbauerManualCoefficients`
        for harmonic in self.harmonic_levels:
            harmonic.gegenbauer = GegenbauerManualCoefficients(
                harmonic.gegenbauer.n, harmonic.gegenbauer.alpha
            )

        max_power = int(max(max(h.gegenbauer.powers) for h in self.harmonic_levels))
        self.num_harmonics = sum(len(h) for h in self.harmonic_levels)

        weights = np.zeros((self.num_harmonics, max_power + 1))  # [M, P]
        powers = np.zeros((self.num_harmonics, max_power + 1))  # [M, P]
        begin = 0
        for harmonic in self.harmonic_levels:
            coeffs = harmonic.gegenbauer.coefficients
            pows = harmonic.gegenbauer.powers
            for c, p in zip(coeffs, pows):
                for i in range(len(harmonic)):
                    weights[begin + i, int(p)] = c
                    powers[begin + i, int(p)] = p
            begin += len(harmonic)

        self.weights = tf.convert_to_tensor(weights)  # [M, P]
        self.powers = tf.convert_to_tensor(powers)  # [M, P]

        self.V = tf.convert_to_tensor(
            np.concatenate([harmonic.V for harmonic in self.harmonic_levels], axis=0)
        )  # [M, D]

        self.L_inv = tf.linalg.LinearOperatorBlockDiag(
            [
                tf.linalg.LinearOperatorFullMatrix(harmonic.L_inv)
                for harmonic in self.harmonic_levels
            ]
        )  # [M, M] block diagonal

    def __call__(self, X: B.Numeric) -> B.Numeric:
        """
        Evaluates each of the spherical harmonics in the collection,
        and stacks the results.
        :param x: TensorType, [N, D]
            N points with unit norm in cartesian coordinate system.
        :return: [num harmonics in collection, N]
        """
        VXT = tf.matmul(self.V, X, transpose_b=True)  # [M, N, 1]
        tmp = self.weights[:, None, :] * (
            VXT[:, :, None] ** self.powers[:, None, :]
        )  # [M, N, P]
        gegenbauer_at_VXT = tf.reduce_sum(tmp, axis=-1)  # [M, N]
        return tf.transpose(self.L_inv.matmul(gegenbauer_at_VXT))  # [N, M]



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib.pyplot as plt
import numpy as np
import plotly.graph_objects as go
from beartype.typing import Callable
from matplotlib import cm


def spherical_grid(resolution=100):
    phi = np.linspace(0, np.pi, resolution)
    theta = np.linspace(0, 2 * np.pi, resolution)
    phi, theta = np.meshgrid(phi, theta)

    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta)
    z = np.cos(phi)

    return np.vstack([x.flatten(), y.flatten(), z.flatten()]).T


def plot_spherical_function(
    f: Callable, resolution: int = 100, rescale_radius: bool = False, ax=None
):
    """
    Use matplotlib to create a 3D plot of a function defined on the sphere.

    :param f: A function which takes a [N, 3] matrix of points on the sphere in
        cartesian coordinates, and returns a sized [N] vector.
    :param resolution: Controls the resolution of the grid (larger means finer grid).
    :param rescale_radius: Slightly modify the radius to depict the function values.
    :param ax: Matplotlib axis to plot. If `None` (default), a new axis is created.
    """

    grid = spherical_grid(resolution)
    fgrid = f(grid).reshape(resolution, resolution)

    # Set the aspect ratio to 1 so our sphere looks spherical
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

    if rescale_radius:
        scale = np.abs(fgrid)
    else:
        scale = 1.0

    # scale the colors
    fmax, fmin = fgrid.max(), fgrid.min()
    fcolors = (fgrid - fmin) / (fmax - fmin)
    viridis = cm.get_cmap("viridis")

    ax.plot_surface(
        grid[:, 0].reshape(resolution, resolution) * scale,
        grid[:, 1].reshape(resolution, resolution) * scale,
        grid[:, 2].reshape(resolution, resolution) * scale,
        rstride=1,
        cstride=1,
        facecolors=viridis(fcolors),
    )

    # Turn off the axis planes
    ax.set_axis_off()
    return ax


def get_mesh_on_sphere(resolution):
    import meshzoo

    grid, cells = meshzoo.icosa_sphere(resolution)
    return grid, cells


def plotly_plot_spherical_function(
    f: Callable, use_mesh: bool = True, resolution: int = 100, animate_steps: int = 0
):
    """
    Use plotly to create a 3D plot of a function defined on the sphere.

    :param f: A function which takes a [N, 3] matrix of points on the sphere in
        cartesian coordinates, and returns a sized [N] vector.
    :param use_mesh: If `True` use a mesh for plotting, otherwise a Surface plot is used.
    :param resolution: Controls the resolution of the mesh (larger means finer mesh).
    :param animate_steps: An animation, in which the sphere is rotating
        `animate_steps` number of steps can be created if setting `animate_steps
        > 0`. If set to 0 (default), no animation is added to the plot.
    """

    if use_mesh:
        grid, cells = get_mesh_on_sphere(resolution)
        x, y, z = [grid[:, i] for i in range(3)]
        fgrid = f(grid)
    else:
        phi = np.linspace(0, np.pi, resolution)
        theta = np.linspace(0, 2 * np.pi, resolution)
        phi, theta = np.meshgrid(phi, theta)

        x = np.sin(phi) * np.cos(theta)
        y = np.sin(phi) * np.sin(theta)
        z = np.cos(phi)

        grid = np.vstack([x.flatten(), y.flatten(), z.flatten()]).T
        fgrid = f(grid).reshape(resolution, resolution)

    # scale the colors
    # fmax, fmin = fgrid.max(), fgrid.min()
    # fcolors = (fgrid - fmin) / (fmax - fmin)
    fcolors = fgrid

    C0 = "rgb(31, 119, 180)"
    C1 = "rgb(255, 127, 14)"
    colorscale = [[0.0, C0], [0.5, "white"], [1, C1]]
    if use_mesh:
        plot = go.Mesh3d(
            x=grid[:, 0],
            y=grid[:, 1],
            z=grid[:, 2],
            i=cells[:, 0],
            j=cells[:, 1],
            k=cells[:, 2],
            colorbar_title="z",
            colorscale=colorscale,
            intensity=fcolors,
        )
    else:
        plot = go.Surface(x=x, y=y, z=z, surfacecolor=fcolors, colorscale=colorscale)
    fig = go.Figure(plot)
    fig.update_layout(scene_aspectmode="cube")
    fig.update_scenes(xaxis_visible=False, yaxis_visible=False, zaxis_visible=False)
    # fig.update_traces(showscale=False, hoverinfo="none")
    fig.update_layout(margin=dict(l=0, r=0, t=0, b=0))

    fig.update_layout(plot_bgcolor="rgba(0,0,0,0)", paper_bgcolor="rgba(0,0,0,0)")
    fig.update_layout(
        scene=dict(
            xaxis=dict(showbackground=False, showticklabels=False, visible=False),
            yaxis=dict(showbackground=False, showticklabels=False, visible=False),
            zaxis=dict(showbackground=False, showticklabels=False, visible=False),
        )
    )

    x_eye = 1
    y_eye = 1
    z_eye = 1
    angle = np.pi / 720

    fig.update_layout(
        width=600,
        height=600,
        scene_camera_eye=dict(x=x_eye, y=y_eye, z=z_eye),
    )

    frames = []

    def rotate(x, y, z):
        r, t, z = xyz2rtz(x, y, z)
        t += angle
        x, y, z = rtz2xyz(r, t, z)
        frames.append(go.Frame(layout=dict(scene_camera_eye=dict(x=x, y=y, z=z))))
        return x, y, z

    def xyz2rtz(x, y, z):
        return (x**2 + y**2) ** 0.5, np.arctan2(y, x), z

    def rtz2xyz(r, t, z):
        return r * np.cos(t), r * np.sin(t), z

    x, y, z = x_eye, y_eye, z_eye

    if animate_steps > 0:
        for i in range(animate_steps):
            print(i)
            x, y, z = rotate(x, y, z)
        fig.frames = frames

    return fig



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import lab as B
import numpy as np
from beartype.typing import List, Tuple, Union
from lab import dispatch
from scipy.special import gegenbauer as scipy_gegenbauer
from scipy.special import loggamma

from spherical_harmonics.lab_extras import polyval


class Polynomial:
    r"""
    One-dimensional polynomial function expressed with coefficients and powers.
    The polynomial f(x) is given by f(x) = \sum_i c_i x^{p_i}, with x \in R^1.
    """

    def __init__(
        self,
        coefficients: Union[List, np.ndarray],
        powers: Union[List, np.ndarray],
    ):
        r"""
        The polynomial f(x) is given by f(x) = \sum_i c_i x^{p_i},
        with c coefficients and p powers. The number of coefficients and
        the number of powers must match.

        :param coefficients: list of weights
        :param powers: list of powers
        """
        assert len(coefficients) == len(powers)
        self.coefficients = coefficients
        self.powers = powers

    def __call__(self, x: B.Numeric) -> B.Numeric:
        """
        Evaluates the polynomial @ `x`
        :param x: 1D input values at which to evaluate the polynomial, [...]

        :return:
            function evaluations, same shape as `x` [...]
        """
        cs = B.reshape(self.coefficients, 1, -1)  # [1, M]
        ps = B.reshape(self.powers, 1, -1)  # [1, M]
        x_flat = B.reshape(x, -1, 1)  # [N, 1]
        val = B.sum(cs * (x_flat**ps), axis=1)  # [N, M]  # [N]
        return B.reshape(val, *B.shape(x))


class GegenbauerManualCoefficients(Polynomial):
    r"""
    Gegenbauer polynomials or ultraspherical polynomials C_n^(α)(x)
    are orthogonal polynomials on the interval [−1,1] with respect
    to the weight function (1 − x^2)^{α–1/2} [1].

    [1] https://en.wikipedia.org/wiki/Gegenbauer_polynomials,
    [2] Approximation Theory and Harmonic Analysis on Spheres and Balls,
        Feng Dai and Yuan Xu, Chapter 1. Spherical Harmonics,
        https://arxiv.org/pdf/1304.2585.pdf
    """

    def __init__(self, n: int, alpha: float):
        """
        Gegenbauer polynomial C_n^(α)(z) of degree `n` and characterisation `alpha`.
        We represent the Gegenbauer function as a polynomial and compute its
        coefficients and corresponding powers.

        :param n: degree
        :param alpha: characterises the form of the polynomial.
            Typically changes with the dimension, alpha = (dimension - 2) / 2
        """

        coefficients, powers = self._compute_coefficients_and_powers(n, alpha)
        super().__init__(
            np.array(coefficients, dtype=np.float64),
            np.array(powers, dtype=np.float64),
        )
        self.n = n
        self.alpha = alpha
        self._at_1 = scipy_gegenbauer(self.n, self.alpha)(1.0)

    def _compute_coefficients_and_powers(
        self, n: int, alpha: float
    ) -> Tuple[List, List]:
        """
        Compute the weights (coefficients) and powers of the Gegenbauer functions
        expressed as polynomial.

        :param n: degree
        :param alpha:
        """
        coefficients, powers = [], []

        for k in range(math.floor(n / 2) + 1):  # k=0 .. floor(n/2) (incl.)
            # computes the coefficients in log space for numerical stability
            log_coef = loggamma(n - k + alpha)
            log_coef -= loggamma(alpha) + loggamma(k + 1) + loggamma(n - 2 * k + 1)
            log_coef += (n - 2 * k) * np.log(2)
            coef = np.exp(log_coef)
            coef *= (-1) ** k
            coefficients.append(coef)
            powers.append(n - 2 * k)

        return coefficients, powers

    def __call__(self, x: B.Numeric) -> B.Numeric:
        if self.n < 0:
            return B.zeros(x)
        elif self.n == 0:
            return B.ones(x)
        elif self.n == 1:
            return 2 * self.alpha * x
        else:
            return super().__call__(x)

    @property
    def value_at_1(self):
        """
        Gegenbauer evaluated at 1.0
        """
        return self._at_1


class GegenbauerScipyCoefficients:
    """Gegenbauer polynomial using the coefficients given by Scipy."""

    def __init__(self, n: int, alpha: float):
        self.n = n
        self.alpha = alpha
        self.C = scipy_gegenbauer(n, alpha)
        self._at_1 = self.C(1.0)
        self.coefficients = list(self.C.coefficients)

    @dispatch
    def __call__(self, x: B.NPNumeric) -> B.Numeric:
        if self.n < 0:
            return B.zeros(x)
        elif self.n == 0:
            return B.ones(x)
        elif self.n == 1:
            return 2 * self.alpha * x

        return self.C(x)

    @dispatch  # type: ignore[no-redef]
    def __call__(self, x: B.Numeric) -> B.Numeric:
        """x: [...], return [...]"""
        if self.n < 0:
            return B.zeros(x)
        elif self.n == 0:
            return B.ones(x)
        elif self.n == 1:
            return 2 * self.alpha * x

        return polyval(self.coefficients, x)

    @property
    def value_at_1(self):
        """Gegenbauer evaluated at 1.0"""
        return self._at_1


Gegenbauer = GegenbauerScipyCoefficients



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from spherical_harmonics import SphericalHarmonics
from spherical_harmonics.plotting import plotly_plot_spherical_function


@pytest.fixture
def spherical_function_to_plot():
    phi = SphericalHarmonics(3, 20)
    func = lambda x: phi(x)[:, -1]
    _ = func(np.random.randn(1, 3))
    return func


# FIXME: the use_mesh=True test is commented out because it relies on meshzoo
# which now requires a paid license. This could be fixed by using, e.g.,
# potpourri3d instead of meshzoo.
# @pytest.mark.parametrize("use_mesh", [True, False])
@pytest.mark.parametrize("use_mesh", [False])
@pytest.mark.parametrize("animate_steps", [0, 11])
def test_plotting(spherical_function_to_plot, use_mesh, animate_steps):
    _ = plotly_plot_spherical_function(
        spherical_function_to_plot,
        resolution=10,
        animate_steps=animate_steps,
        use_mesh=use_mesh,
    )



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Spherical Harmonics and associated utility functions"""

import lab as B
import numpy as np
from beartype.typing import List, Union
from scipy.special import gegenbauer as scipy_gegenbauer

from spherical_harmonics.fundamental_set import FundamentalSystemCache, num_harmonics
from spherical_harmonics.gegenbauer_polynomial import Gegenbauer
from spherical_harmonics.lab_extras import from_numpy
from spherical_harmonics.utils import surface_area_sphere


class SphericalHarmonics:
    r"""
    Contains all the spherical harmonic levels up to a `max_degree`.
    Total number of harmonics in the collection is given by
    :math:`\sum_{degree=0}^{max_degree-1} num_harmonics(dimension, degree)`
    """

    def __init__(
        self,
        dimension: int,
        degrees: Union[int, List[int]],
        debug: bool = False,
        allow_uncomputed_levels: bool = False,
    ):
        """
        :param dimension: if d = dimension, then
            S^{d-1} = { x ∈ R^d and ||x||_2 = 1 }
            For a circle d=2, for a ball d=3
        :param degrees: list of degrees of spherical harmonic levels,
            if integer all levels (or degrees) up to `degrees` are used.
        highest degree of polynomial
            in the collection (exclusive)
        :param allow_uncomputed_levels: if True, allow levels without the precomputed
            fundamental system.
        :param debug: print debug messages.
        """
        assert (
            dimension >= 3
        ), f"Lowest supported dimension is 3, you specified {dimension}"
        self.debug = debug

        if isinstance(degrees, int):
            degrees = list(range(degrees))

        self.fundamental_system = FundamentalSystemCache(
            dimension, strict_loading=not allow_uncomputed_levels
        )
        self.harmonic_levels = [
            SphericalHarmonicsLevel(dimension, degree, self.fundamental_system)
            for degree in degrees
        ]

    def __call__(
        self,
        x: B.Numeric,
    ) -> B.Numeric:
        """
        Evaluates each of the spherical harmonic level in the collection,
        and stacks the results.
        :param x: [N, D]
            N points with unit norm in cartesian coordinate system.
        :return: [N, num harmonics in collection]
        """
        values = map(
            lambda harmonic: harmonic(x), self.harmonic_levels
        )  # List of length `max_degree` with Tensor [num_harmonics_degree, N]

        return B.transpose(B.concat(*list(values), axis=0))  # [num_harmonics, N]

    def __len__(self):
        return sum(len(harmonic_level) for harmonic_level in self.harmonic_levels)

    def num_levels(self):
        return len(self.harmonic_levels)

    def addition(self, X, X2=None):
        """For test purposes only"""
        return B.sum(
            B.stack(
                *[level.addition(X, X2) for level in self.harmonic_levels],
                axis=0,
            ),
            axis=0,
        )  # [N1, N2]


class SphericalHarmonicsLevel:
    r"""
    Spherical harmonics \phi(x) in a specific dimension and degree (or level).

    The harmonics are constructed by
    1) Building a fundamental set of directions {v_i}_{i=1}^N,
        where N is number of harmonics of the degree.
        Given these directions we have that {c(<v_i, x>)}_i is a basis,
        where c = gegenbauer(degree, alpha) and alpha = (dimension - 2)/2.
        See Definition 3.1 in [1].
    2) Using Gauss Elimination we orthogonalise this basis, which
       corresponds to the Gram-Schmidt procedure.

    [1] Approximation Theory and Harmonic Analysis on Spheres and Balls,
        Feng Dai and Yuan Xu, Chapter 1. Spherical Harmonics,
        https://arxiv.org/pdf/1304.2585.pdf
    """

    def __init__(self, dimension: int, degree: int, fundamental_system=None):
        r"""
        param dimension: if d = dimension, then
            S^{d-1} = { x ∈ R^d and ||x||_2 = 1 }
            For a circle d=2, for a ball d=3
        param degree: degree of the harmonic, also referred to as level.
        """
        assert (
            dimension >= 3
        ), f"Lowest supported dimension is 3, you specified {dimension}"
        self.dimension, self.degree = dimension, degree
        self.alpha = (self.dimension - 2) / 2.0
        self.num_harmonics_in_level = num_harmonics(self.dimension, self.degree)

        self.V = fundamental_system.load(self.degree)

        # surface area of S^{d−1}
        self.surface_area_sphere = surface_area_sphere(dimension)
        # normalising constant
        c = self.alpha / (degree + self.alpha)

        if self.is_level_computed:
            VtV = np.dot(self.V, self.V.T)
            self.A = c * scipy_gegenbauer(self.degree, self.alpha)(VtV)
            self.L = np.linalg.cholesky(self.A)  # [M, M]
            # Cholesky inverse corresponds to the weights you get from Gram-Schmidt
            self.L_inv = np.linalg.solve(self.L, np.eye(len(self.L)))
        self.gegenbauer = Gegenbauer(self.degree, self.alpha)

    @property
    def is_level_computed(self) -> bool:
        """
        Whether the level has the fundamental system computed.
        """
        return self.V is not None

    def __call__(self, X: B.Numeric) -> B.Numeric:
        r"""
        :param X: M normalised (i.e. unit) D-dimensional vector, [N, D]

        :return: `X` evaluated at the M spherical harmonics in the set.
            [\phi_m(x_i)], shape [M, N]
        """
        if not self.is_level_computed:
            raise ValueError(
                f"Fundamental system for dimension {self.dimension} and degree "
                f"{self.degree} has not been precomputed. Terminating "
                "computations. Precompute set by running `fundamental_set.py`"
            )

        VXT = B.matmul(
            B.cast(B.dtype(X), from_numpy(X, self.V)), X, tr_b=True
        )  # [M, N]
        zonals = self.gegenbauer(VXT)  # [M, N]
        return B.matmul(B.cast(B.dtype(X), self.L_inv), zonals)  # [M, N]

    # TODO(Vincent) for some reason Optional[B.Numeric] doesn't work
    def addition(self, X: B.Numeric, Y: B.Numeric = None) -> B.Numeric:
        r"""
        Addition theorem. The sum of the product of all the spherical harmonics evaluated
        at x and x' of a specific degree simplifies to the gegenbauer polynomial evaluated
        at the inner product between x and x'.

        Mathematically:
            \sum_{k=1}^{N(dim, degree)} \phi_k(X) * \phi_k(Y)
                = (degree + \alpha) / \alpha * C_degree^\alpha(X^T Y)
        where \alpha = (dimension - 2) / 2 and omega_d the surface area of the
        hypersphere S^{d-1}.

        :param X: Unit vectors on the (hyper) sphere [N1, D]
        :param Y: Unit vectors on the (hyper) sphere [N2, D].
            If None, X is used as Y.

        :return: [N1, N2]
        """
        if Y is None:
            Y = X
        XYT = B.matmul(X, Y, tr_b=True)  # [N1, N2]
        c = self.gegenbauer(XYT)  # [N1, N2]
        return (self.degree / self.alpha + 1.0) * c  # [N1, N2]

    def addition_at_1(self, X: B.Numeric) -> B.Numeric:
        r"""
        Evaluates \sum_k \phi_k(x) \phi_k(x), notice the argument at which we evaluate
        the harmonics is equal. See `self.addition` for the general case.

        This simplifies to
            \sum_{k=1}^{N(dim, degree)} \phi_k(x) * \phi_k(x)
                = (degree + \alpha) / \alpha * C_degree^\alpha(1)

        as all vectors in `X` are normalised so that x^\top x == 1.

        :param X: only used for it's X.shape[0], [N, D]
        :return: [N, 1]
        """
        c = (
            B.ones(
                X.dtype,
                *(X.shape[0], 1),
            )
            * self.gegenbauer.value_at_1
        )  # [N, 1]
        return (self.degree / self.alpha + 1.0) * c  # [N, 1]

    def eigenvalue(self) -> Union[int, float, np.ndarray]:
        """
        Spherical harmonics are eigenfunctions of the Laplace-Beltrami operator
        (also known as the Spherical Laplacian). We return the associated
        eigenvalue.

        The eigenvalue of the N(dimension, degree) number of spherical harmonics
        on the same level (i.e. same degree) is the same.
        """
        return eigenvalue_harmonics(self.degree, self.dimension)

    def __len__(self):
        return self.num_harmonics_in_level


def eigenvalue_harmonics(
    degrees: Union[int, float, np.ndarray], dimension: int
) -> Union[int, float, np.ndarray]:
    """
    Eigenvalue of a spherical harmonic of a specific degree.

    :param degrees: a single, or a array of degrees
    :param dimension:
        S^{d-1} = { x ∈ R^d and ||x||_2 = 1 }
        For a circle d=2, for a ball d=3

    :return: the corresponding eigenvalue of the spherical harmonic
        for the specified degrees, same shape as degrees.
    """
    assert dimension >= 3, "We only support dimensions >= 3"

    return degrees * (degrees + dimension - 2)



import lab as B
import torch
from beartype.typing import List
from lab import dispatch
from plum import Union

_Numeric = Union[B.Number, B.TorchNumeric]


@dispatch
def polyval(coeffs: list, x: _Numeric) -> _Numeric:  # type: ignore
    """
    Computes the elementwise value of a polynomial.

    If `x` is a tensor and `coeffs` is a list if size n + 1, this function returns
    the value of the n-th order polynomial

    ..math:
        p(x) = coeffs[n-1] + coeffs[n-2] * x + ... + coeffs[0] * x**(n-1)
    """
    curVal = 0
    for i in range(len(coeffs) - 1):
        curVal = (curVal + coeffs[i]) * x

    return curVal + coeffs[-1]


@dispatch
def from_numpy(
    a: B.TorchNumeric, b: Union[List, B.Number, B.NPNumeric, B.TorchNumeric]
):  # type: ignore
    """
    Converts the array `b` to a tensor of the same backend as `a`
    """
    return torch.tensor(b)



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lab as B
import numpy as np
from scipy.special import gamma


def surface_area_sphere(d: int) -> float:
    r"""
    Surface area of sphere in R^d, S^{d-1}.

    \Omega_{d-1} following notation of [1].

    :param d: dimension
        S^{d-1} = { x ∈ R^d and ||x||_2 = 1 }
        For a ball d=3

    [1] Sparse Gaussian Processes with Spherical Harmonic Features,
        Vincent Dutordoir, Nicolas Durrande and James Hensman.
        Appendix, equation 2.
    """
    return 2 * (np.pi ** (d / 2)) / gamma(d / 2)


def spherical_to_cartesian_4d(thetas, r=1.0):
    """
    Converts points on the hypersphere S^3 in R^4 given in
    spherical coordinates to their cartesian value.

    :param thetas: [N, 3]
        thetas = [theta_1 | theta_2 | theta_3], with theta_i [N, 1]
        0 <= theta_1 < 2 * pi and
        0 <= theta_2 < pi
        0 <= theta_3 < pi
    :return: points in cartesian coordinate system [N, 4],
        norm of the points equals `r`.
    """
    assert thetas.shape[-1] == 3

    theta1, theta2, theta3 = [thetas[:, [i]] for i in range(3)]
    x1 = r * np.sin(theta3) * np.sin(theta2) * np.sin(theta1)
    x2 = r * np.sin(theta3) * np.sin(theta2) * np.cos(theta1)
    x3 = r * np.sin(theta3) * np.cos(theta2)
    x4 = r * np.cos(theta3)
    return np.c_[x1, x2, x3, x4]


def spherical_to_cartesian(thetas, r=1.0):
    """
    Converts points on the hypersphere S^3 in R^4 given in
    spherical coordinates to their cartesian value.

    :param thetas: [N, 2]
        thetas = [theta_1 | theta_2], with theta_i [N, 1]
        0 <= theta_1 < 2 * pi and
        0 <= theta_2 < pi
    :return: points in cartesian coordinate system [N, 3],
        norm of the points equals `r`.
    """
    assert thetas.shape[-1] == 2

    theta = thetas[:, [0]]
    phi = thetas[:, [1]]
    x = r * np.sin(phi) * np.cos(theta)
    y = r * np.sin(phi) * np.sin(theta)
    z = r * np.cos(phi)
    return np.c_[x, y, z]


def l2norm(X: B.Numeric) -> B.Numeric:
    """
    Returns the norm of the vectors in `X`. The vectors are
    D-dimensional and  stored in the last dimension of `X`.

    :param X: [..., D]
    :return: norm for each element in `X`, [N, 1]
    """
    return B.sum(X**2, squeeze=False, axis=-1) ** 0.5



# Copyright 2021 Vincent Dutordoir. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

requirements = [
    "numpy",
    "scipy",
]

setup(
    name="Spherical Harmonics",
    version="0.0.2a0",
    author="vd309@cam.ac.uk",
    description="Python Implementation of Spherical harmonics in dimension >= 3",
    packages=find_packages("src"),  # include all packages under src
    package_dir={"": "src"},  # tell distutils packages are under src
    install_requires=requirements,
    include_package_data=True,
    package_data={"spherical_harmonics": ["fundamental_system/*.npz"]},
)



import operator
from functools import reduce

import numpy as np
import pytest
import tensorflow as tf
from scipy.integrate import quad
from scipy.special import factorial, gamma
from scipy.special import gegenbauer as scipy_gegenbauer

import spherical_harmonics.tensorflow  # noqa
from spherical_harmonics.gegenbauer_polynomial import (
    Gegenbauer,
    GegenbauerScipyCoefficients,
    Polynomial,
)
from spherical_harmonics.spherical_harmonics import num_harmonics
from spherical_harmonics.utils import surface_area_sphere


def test_polynomial():
    x = np.random.rand(5, 6)
    cs = np.array([2.0, 4.0, 3.5], dtype=np.float64)
    ps = np.array([3.0, 5.0, 8.0], dtype=np.float64)

    np.testing.assert_array_almost_equal(
        reduce(operator.add, (c * x**p for c, p in zip(cs, ps))),
        Polynomial(cs, ps)(x),
    )

    np.testing.assert_array_almost_equal(
        cs[0] * x ** ps[0] + cs[1] * x ** ps[1] + cs[2] * x ** ps[2],
        Polynomial(cs, ps)(x),
    )


@pytest.mark.parametrize("alpha", [0.5, 1.5, 2.0, 9.0])
@pytest.mark.parametrize("n", range(10))
@pytest.mark.parametrize("GegenbauerClass", [Gegenbauer, GegenbauerScipyCoefficients])
def test_Gegenbauer(alpha, n, GegenbauerClass):
    x = np.linspace(-1, 1, 1_000).reshape(-1, 1)

    np.testing.assert_array_almost_equal(
        scipy_gegenbauer(n, alpha)(x),
        GegenbauerClass(n, alpha)(tf.convert_to_tensor(x)),
    )


@pytest.mark.parametrize("max_degree,dimension", [(8, 10), (7, 20), (6, 31)])
def test_Gegenbauer_extreme(max_degree, dimension):
    alpha = (dimension - 2) / 2
    x = np.linspace(-1, 1, 1_000).reshape(-1, 1)

    np.testing.assert_array_almost_equal(
        scipy_gegenbauer(max_degree, alpha)(x),
        Gegenbauer(max_degree, alpha)(tf.convert_to_tensor(x)),
    )

    np.testing.assert_array_almost_equal(
        scipy_gegenbauer(max_degree, alpha)(x),
        GegenbauerScipyCoefficients(max_degree, alpha)(tf.convert_to_tensor(x)),
    )


@pytest.mark.parametrize("degree", range(8))
@pytest.mark.parametrize("dimension", range(3, 9))
@pytest.mark.parametrize("GegenbauerClass", [Gegenbauer, GegenbauerScipyCoefficients])
def test_normalisation_gegenbauer(degree, dimension, GegenbauerClass):
    omega_d = surface_area_sphere(dimension) / surface_area_sphere(dimension - 1)
    alpha = (dimension - 2) / 2
    gegenbauer = GegenbauerClass(degree, alpha)

    def c(t):
        return gegenbauer(tf.cast(t, dtype=tf.float64)).numpy()

    def func(t):
        return c(t) ** 2 * (1 - t**2) ** (alpha - 0.5)

    desired = quad(func, -1, 1)[0]
    value = c(1) * alpha / (degree + alpha) * omega_d
    np.testing.assert_almost_equal(desired, value, decimal=5)

    # Definition Wiki https://en.wikipedia.org/wiki/Gegenbauer_polynomials
    value2 = np.pi * 2 ** (1 - 2 * alpha) * gamma(degree + 2 * alpha)
    value2 /= factorial(degree) * (degree + alpha) * gamma(alpha) ** 2
    np.testing.assert_almost_equal(desired, value2, decimal=5)


@pytest.mark.parametrize("alpha", [0.5, 1.5, 2.0, 9.0])
@pytest.mark.parametrize("n", range(10))
@pytest.mark.parametrize("GegenbauerClass", [Gegenbauer, GegenbauerScipyCoefficients])
def test_gegenbauer_at_1(n, alpha, GegenbauerClass):
    c_1 = GegenbauerClass(n, alpha)(tf.cast(1.0, dtype=tf.float64)).numpy()
    expected = gamma(2 * alpha + n) / gamma(2 * alpha) / factorial(n)
    np.testing.assert_almost_equal(c_1, expected)

    # another test
    dim = int(2 * alpha + 2)
    value1 = c_1 * (n + alpha) / alpha
    N = num_harmonics(dim, n)
    assert round(value1) == N
